/** \file cprolog.h
 *
 * client-side specific macros ...
 *
 */
#ifndef CPROLOG_H_
#define CPROLOG_H_

#include "project.def"
#include "prolog.h"
#include "errors.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

void closeIPClients(void);
#if defined(IPXNW) || defined(IPXBSD)
   void closeIPXClients(void);
   int initIPXClient(int socket);
   int sendSPXData(WORD ID,int socket);
#  define ShutdownClients(i) {\
     closeIPXClients();\
     closeIPClients();\
   }
#  define ConsumerCheck(i) {\
     if (i == 0) {ReturnToServerCycle++; saveClientCycle = ReturnToClientCycle;}\
     doIPXConsumer();\
     doUDPConsumer();\
     if (i == 0) {ReturnToServerCycle--; ReturnToClientCycle = saveClientCycle;}\
   }
#else
#  define sendSPXData(i,j) ipx_not_supported
#  define ShutdownClients(i) {\
	   closeIPClients();\
   }
#  define ConsumerCheck(i) {\
     if (i == 0) {ReturnToServerCycle++; saveClientCycle = ReturnToClientCycle;}\
     doUDPConsumer();\
     if (i == 0) {ReturnToServerCycle--; ReturnToClientCycle = saveClientCycle;}\
   }
#endif  /* IPXNW || IPXBSD */

#ifdef DOS   /* definitions */
#  define clnWorkArea     WorkArea
#  define clnWorkAreaSize WorkAreaSize
#  define whoami(c) ;
#else
   extern BYTE *clnWorkArea;
   extern UINT32 clnWorkAreaSize;
   char *whoami(char *user);
   void ClientDelay(int sec);  /* delay(0) never returns */
#endif

/**
 * \internal
 * \brief Increments the connection table
 * 
 * Trivial macro wrapper which is the same as n++.
 * 
 * \todo consider ways of dynamically incrementing connection table past MaxNumConnections
 * (at least for some platforms).
 */
#define IncrementConnections(n) (n)++
/**
 * \internal
 * \brief Empty macro
 * 
 * A place-holder in clnip.c and clnipx.c, in tandem with ReleaseWorkArea().  As 
 * client cycles and server cycles generally use different work areas on all OSes
 * capable of using threads, we should never have to worry about synchronizing 
 * a common work area.  Note that DOS and NIOS both use a common work area to save
 * memory but are inherently single threaded.
 * 
 * \todo get rid of these lines of code.
 * 
 */
#define WaitForWorkArea(i) ;
/**
 * \internal
 * \brief Empty macro
 * 
 * A place-holder in clnip.c and clnipx.c, in tandem with WaitForWorkArea().  As 
 * client cycles and server cycles generally use different work areas on all OSes
 * capable of using threads, we should never have to worry about synchronizing 
 * a common work area.  Note that DOS and NIOS both use a common work area to save
 * memory but are inherently single threaded.
 * 
 * \todo get rid of these lines of code.
 * 
 */
#define ReleaseWorkArea(i) ;

/**
 * Allocates memory for a connection table entry
 *
 * \param n Index in connection table
 * \todo replace macro by a function.
 * \todo ??? means int addConnectionTableEntry() { allocate, increment_counter }
 * \bug does not increment counter (see getConTblEntry())
 */
#define AssignConnTableMemory(n) \
  ( n >= MaxNumConnections ? out_of_client_memory :\
    (conTbl[n]=(ConTblEntry *)SystemCalloc(1,sizeof(ConTblEntry))) == NULL ? out_of_client_memory :\
   0 )
#define CheckSubscriptions(e) \
   while (prepSubRequest(clnWorkArea,e) >= 0);
#define prepNameServerEntry(i) {\
   conTbl[i] = &connNameServer;\
   strcpy(conTbl[i]->devName,devnam);\
   strcpy(conTbl[i]->prpName,devprp);\
   if (dout != NULL) {\
     conTbl[i]->sizeOut = dout->dArrayLength;\
     conTbl[i]->formatOut = dout->dFormat;\
     conTbl[i]->dataOut = (char *)dout->data.vptr;\
   } else {\
     conTbl[i]->sizeOut = 0;\
     conTbl[i]->formatOut = CF_NULL;\
     conTbl[i]->dataOut = NULL;\
   }\
   if (din != NULL) {\
     conTbl[i]->sizeIn = din->dArrayLength;\
     conTbl[i]->formatIn = din->dFormat;\
     conTbl[i]->dataIn = (char *)din->data.vptr;\
     conTbl[i]->BytesTotal = conTbl[i]->sizeIn*fmtsizeof(conTbl[i]->formatIn);\
     conTbl[i]->datamtu = UDP_BUFFER_SIZE-SUBSCRIPTION_SIZE-HEADERSTRUCT_SIZE;\
     if (conTbl[i]->BytesTotal < conTbl[i]->datamtu) {\
       conTbl[i]->numblksIn = 1;\
       conTbl[i]->BytesRemaining = conTbl[i]->BytesTotal;\
       conTbl[i]->nextDataSize = conTbl[i]->BytesTotal;\
     } else { conTbl[i]->BytesTotal = 0; }\
   } else {\
     conTbl[i]->sizeIn = 0;\
     conTbl[i]->formatIn = CF_NULL;\
     conTbl[i]->dataIn = NULL;\
   }\
   conTbl[i]->devNum = devNum;\
   conTbl[i]->notifyPending = 1;\
   if (conTbl[i]->bufOut) { SystemFree(conTbl[i]->bufOut); conTbl[i]->bufOut = NULL; }\
   if (conTbl[i]->bufOutBlks) { SystemFree(conTbl[i]->bufOutBlks); conTbl[i]->bufOutBlks = NULL; }\
   if (conTbl[i]->bufCmp) { SystemFree(conTbl[i]->bufCmp); conTbl[i]->bufCmp = NULL; }\
   if (conTbl[i]->dtf) { SystemFree(conTbl[i]->dtf); conTbl[i]->dtf = NULL; }\
}

#define WaitForSyncCount(i) clientCycle()

#if defined(DOS)             /* DOS: initTransport */
#  define initTransport(i)\
   ( (cc=initIPXClient(i)) != 0 && cc != ipx_not_supported ? cc :\
     (cc=initIPClient(i))  != 0 && cc != tcp_not_supported ? cc : 0)
#else                         /* NOT DOS  */
/**
 * \internal
 * \brief Initializes clnWorkArea
 *
 * Allocates memory for the input buffer.
 * \todo implement as a function
 * \todo handle DOS specials in that function - don't use overwriting #defines
 */
#  define initClnWorkArea(i) \
   ( (clnWorkArea=(BYTE *)SystemCalloc(1,clnWorkAreaSize=i)) == NULL ?\
     out_of_client_memory : 0 )
#  ifdef VXWORKS              /* VXWORKS */
     void tineClientCycle(void);
#    define initConsumerProcess(c) \
     (taskSpawn(c,100,0,10000,tineClientCycle,0,0,0,0,0,0,0,0,0,0) == ERROR ?\
      not_running : 0)
#    define initTransport(i)\
     ( (cc=initControlStructs()) ? cc :\
       (cc=initIPClient(i)&~tcp_not_supported) ? cc :\
       (cc=initClnWorkArea(clnWorkAreaSize))\
     )
#  else                       /* NOT VXWORKS */
/**
 * \internal
 * \brief Initializes transport protocols
 *
 * This macro depends on the operating system.
 * It calls initControlStructs() (NOT for DOS), initIPClient(),
 * initIPXClient (only for DOS or idf IPXBSD is defined),
 * initClnWorkArea() (NOT for DOS).
 *
 * \todo Better implement as a function.
 */
#    ifdef IPXBSD
#    define initTransport(i)\
     ( (cc=initControlStructs()) ? cc :\
       (cc=initIPClient(i)&~tcp_not_supported) ? cc :\
       (cc=initIPXClient(i)&~ipx_not_supported) ? cc :\
       initClnWorkArea(clnWorkAreaSize) )
#    else                     /* NO IPX */
#    define initTransport(i)\
     ( (cc=initControlStructs()) ? cc :\
       (cc=initIPClient(i)&~tcp_not_supported) ? cc :\
       initClnWorkArea(clnWorkAreaSize) )
#    endif
#  endif
#endif                       /* DOS initTransport */

#define CompIncomingData(out,cmp,sz,f,t) (\
     cmp == NULL ? FALSE :\
     t   != NULL ? !_datcmp(out,cmp,sz,f,t) :\
     !memcmp(out,cmp,sz)\
  )
#ifdef INTEL  /* data preparation */
#  define AssignClnStructDataPtr(dst,src,dir) {\
     dst = (BYTE *)src;\
   }
#  define cpyClnStructData(dst,src,siz) {\
     memcpy(conTbl[slot]->dataIn,din->data.vptr,siz);\
   }
#  define AssignOutgoingHeader(h,buf) {\
     if (!ForceStructAlignment) h = (HeaderStruct *)buf;\
     memset(h,0,sizeof(HeaderStruct));\
   }
#  define AssignOutgoingBrdcstData(h,buf) h = (GlobalDataHdr *)buf; memset(h,0,sizeof(GlobalDataHdr));
#  define PrepIncomingData(dst,src,dsiz,rsiz,fmt) {\
     if (fmt == CF_STRUCT && c != NULL && (s=getStructFromTag(c->strTagOut))) {\
       StructSwapEx(c->strTagOut,(BYTE *)src,(BYTE *)dst,RTOL,(int)rsiz/s->rawlen,s->size);\
     } else {\
       memcpy(dst,src,MIN(dsiz,rsiz));\
       if ((dsiz) > (rsiz)) memset(&((char *)(dst))[rsiz],0,dsiz-rsiz);\
     }\
     if (c != NULL && c->field[0] != 0) mapBitfieldToTarget(c->context,c->expName,c->strTagOut,c->field,fmt,dst);\
   }
#  define PrepOutgoingHeader(h,dat) {\
     if (ForceStructAlignment) memcpy(dat,h,HEADERSTRUCT_SIZE);\
   }
#  define PrepOutgoingGCastData(h,dat) ;
#  define PrepIncomingGCastData(h,dat) h = (GlobalDataHdr *)dat;
#  define PrepOutgoingData(d,s,siz,fmt) memcpy(d,s,siz)
#else
#  define AssignClnStructDataPtr(dst,src,dir) {\
     if (dir == RTOL) {\
       StructSwapEx(din->dTag,(BYTE *)src,(BYTE *)s->sdata,RTOL,din->dArrayLength,s->size);\
       dst = s->sdata;\
     }\
   }
#  define cpyClnStructData(dst,src,siz) {\
     if (din->dFormat == CF_STRUCT && (s=getStructFromTag(din->dTag)) != NULL)\
       StructSwapEx(din->dTag,(BYTE *)dst,(BYTE *)src,LTOR,din->dArrayLength,s->size);\
     else memcpy(conTbl[slot]->dataIn,din->data.vptr,siz);\
   }
#  define AssignOutgoingHeader(h,buf) memset(h,0,sizeof(HeaderStruct))
#  define AssignOutgoingBrdcstData(h,buf) memset(h,0,sizeof(GlobalDataHdr))
#  define PrepOutgoingHeader(h,dat) {\
     _StructSwap(HeaderStructStruct,(unsigned char *)dat,(unsigned char *)h,LTOR);\
   }
#  define PrepOutgoingGCastData(h,dat) {\
     _StructSwap(GlobalDataHdrStruct,(unsigned char *)dat,(unsigned char *)h,LTOR);\
   }
#  define PrepIncomingGCastData(h,dat) {\
     _StructSwap(GlobalDataHdrStruct,(unsigned char *)dat,(unsigned char *)h,RTOL);\
   }
#  define PrepIncomingData(dst,src,dsiz,rsiz,fmt) {\
     if (fmt == CF_STRUCT && c != NULL && (s=getStructFromTag(c->strTagOut))) {\
       StructSwapEx(c->strTagOut,(BYTE *)src,(BYTE *)dst,RTOL,rsiz/s->rawlen,s->size);\
     } else {\
       memcpy(dst,src,MIN(dsiz,rsiz));\
       SwapBytes((char *)dst,rsiz/getFormatSize(fmt),fmt,RTOL);\
     }\
     if ((dsiz) > (rsiz)) memset(&((char *)(dst))[rsiz],0,dsiz-rsiz);\
     if (c->field[0] != 0) mapBitfieldToTarget(c->context,c->expName,c->strTagOut,c->field,fmt,dst);\
   }
#  define PrepOutgoingData(d,s,siz,fmt) {\
     memcpy(d,s,siz);\
     SwapBytes((char *)d,siz/getFormatSize(fmt),fmt,LTOR);\
   }

#endif  /* INTEL data preparation */

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* CPROLOG_H_ */

